@@ -39,6 +39,7 @@ from torch import nn
 from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence, pad_sequence
 
 from hopsparser import deptree, lexers
+from hopsparser import utils
 from hopsparser.deptree import DepGraph
 from hopsparser.lexers import (
     BertLexer,
@@ -236,7 +237,6 @@ _T_BiAffineParser = TypeVar("_T_BiAffineParser", bound="BiAffineParser")
 class BiAffineParser(nn.Module):
     """Biaffine Dependency Parser."""
 
-    PAD_IDX: Final[int] = 0
     PAD_TOKEN: Final[str] = "<pad>"
     UNK_WORD: Final[str] = "<unk>"
     # Labels that are -100 are ignored in torch crossentropy (we still set it explicitely)
@@ -433,7 +433,7 @@ class BiAffineParser(nn.Module):
         overall_size = torch.zeros(1, dtype=torch.long, device=device)
         gloss = torch.zeros(1, dtype=torch.float, device=device)
 
-        with torch.no_grad():
+        with torch.inference_mode():
             for batch in dev_set:
                 overall_size += batch.sentences.content_mask.sum()
 
@@ -505,13 +505,14 @@ class BiAffineParser(nn.Module):
 
     def train_model(
         self,
-        train_set: "DependencyDataset",
         epochs: int,
         lr: float,
         lr_schedule: LRSchedule,
         model_path: Union[str, pathlib.Path],
+        train_set: "DependencyDataset",
         batch_size: Optional[int] = None,
         dev_set: Optional["DependencyDataset"] = None,
+        log_epoch: Callable[[str, Dict[str, str]], Any] = lambda x, y: None,
     ):
         model_path = pathlib.Path(model_path)
         weights_file = model_path / "weights.pt"
@@ -569,7 +570,7 @@ class BiAffineParser(nn.Module):
                     tagger_scores, arc_scores, lab_scores, batch, loss_fnc
                 )
 
-                with torch.no_grad():
+                with torch.inference_mode():
                     train_loss += loss
 
                 optimizer.zero_grad()
@@ -584,13 +585,15 @@ class BiAffineParser(nn.Module):
                     ),
                     batch_size=batch_size,
                 )
-                logger.info(
-                    f"Epoch {e}"
-                    f" train loss {train_loss.true_divide(overall_size).item():.4f}"
-                    f" dev loss {dev_loss:.4f}"
-                    f" dev tag acc {dev_tag_acc:.2%}"
-                    f" dev head acc {dev_arc_acc:.2%}"
-                    f" dev deprel acc {dev_lab_acc:.2%}"
+                log_epoch(
+                    str(e),
+                    {
+                        "train loss": f"{train_loss.true_divide(overall_size).item():.4f}",
+                        "dev loss": f"{dev_loss:.4f}",
+                        "dev tag acc": f"{dev_tag_acc:06.2%}",
+                        "dev head acc": f"{dev_arc_acc:06.2%}",
+                        "dev deprel acc": f"{dev_lab_acc:06.2%}",
+                    },
                 )
 
                 if dev_arc_acc > best_arc_acc:
@@ -733,7 +736,7 @@ class BiAffineParser(nn.Module):
         self.eval()
         device = next(self.parameters()).device
 
-        with torch.no_grad():
+        with torch.inference_mode():
             for batch in batch_lst:
                 batch = batch.to(device)
 
@@ -901,7 +904,6 @@ class BiAffineParser(nn.Module):
                     unk_word=cls.UNK_WORD,
                     word_dropout=lexer_config["word_dropout"],
                     words=(DepGraph.ROOT_TOKEN, cls.UNK_WORD, *corpus_words),
-                    words_padding_idx=cls.PAD_IDX,
                 )
             elif lexer_config["type"] == "chars_rnn":
                 lexer = CharRNNLexer.from_chars(
@@ -1020,13 +1022,6 @@ class DependencyDataset:
     ):
         self.parser = parser
         self.treelist = treelist
-
-        self.itolab = self.parser.labels
-        self.labtoi = {label: idx for idx, label in enumerate(self.itolab)}
-
-        self.itotag = self.parser.tagset
-        self.tagtoi = {tag: idx for idx, tag in enumerate(self.itotag)}
-
         self.encoded_trees: List[EncodedTree] = []
 
     def encode(self):
@@ -1067,11 +1062,12 @@ def train(
     config_file: pathlib.Path,
     model_path: pathlib.Path,
     train_file: pathlib.Path,
+    dev_file: Optional[pathlib.Path] = None,
     device: Union[str, torch.device] = "cpu",
     max_tree_length: Optional[int] = None,
+    log_epoch: Callable[[str, Dict[str, str]], Any] = utils.log_epoch,
     overwrite: bool = False,
     rand_seed: Optional[int] = None,
-    dev_file: Optional[pathlib.Path] = None,
 ):
     if rand_seed is not None:
         random.seed(rand_seed)
@@ -1130,6 +1126,7 @@ def train(
         batch_size=hp["batch_size"],
         dev_set=devset,
         epochs=hp["epochs"],
+        log_epoch=log_epoch,
         lr=lr_config["base"],
         lr_schedule=lr_config.get(
             "schedule", {"shape": "exponential", "warmup_steps": 0}
